from usttc.exceptions.exceptions import ConfigurationException


class Config:
    def __init__(self, language: str = "en-US", diarization: tuple = None, separate_speaker_per_channel: bool = False):
        """
        :param language:
        :param diarization: To enable diarization, set to tuple (min_speaker_count, max_speaker_count)
        :param separate_speaker_per_channel:
        """
        self.language = language
        if diarization and separate_speaker_per_channel:
            raise ConfigurationException(
                "Cannot configure diarization and separate_recognition_per_channel together. "
                "If you want to detect multiple speakers from mono audio, please configure diarization. "
                "If the audio has multiple channels, please use separate_speaker_per_channel"
            )
        self.diarization = None
        if diarization:
            if (len(diarization) != 2) or \
                    (not isinstance(diarization[0], int)) or (not isinstance(diarization[1], int)) or \
                    (diarization[0] > diarization[1]) or (diarization[0] < 1):
                raise ConfigurationException("Diarization config is invalid")
            if diarization[1] > 1:
                self.diarization = diarization
        self.separate_speaker_per_channel = separate_speaker_per_channel



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.asr_client.asr_provider import AsrProvider
import requests
import time


class AssemblyAiClient(AsrClient):
    provider = AsrProvider.ASSEMBLY_AI

    def __init__(self, token):
        self.token = token

    def recognize(self, audio: AudioFile, config: Config = Config()):

        def read_file(filename, chunk_size=5242880):
            with open(filename, 'rb') as _file:
                while True:
                    data = _file.read(chunk_size)
                    if not data:
                        break
                    yield data

        headers = {'authorization': self.token}
        response = requests.post('https://api.assemblyai.com/v2/upload',
                                 headers=headers,
                                 data=read_file(audio.file))
        audio_url = response.json()["upload_url"]
        endpoint = "https://api.assemblyai.com/v2/transcript"
        dual_channel = config.separate_speaker_per_channel and (audio.channels == 2)
        json = {
            "audio_url": audio_url,
            "speaker_labels": (config.diarization is not None),
            "dual_channel": dual_channel
        }

        response = requests.post(endpoint, json=json, headers=headers)

        _id = response.json()["id"]
        while True:
            time.sleep(5)
            polling_response = requests.get("https://api.assemblyai.com/v2/transcript/" + _id, headers=headers)
            r = polling_response.json()
            if r['status'] == 'completed':
                transcript = r["text"]
                words = []
                for w in r["words"]:
                    if dual_channel:
                        speaker = w.get("channel")
                    else:
                        speaker = w.get("speaker")
                    words.append(
                        Word(text=w["text"], confidence=w["confidence"], start=w["start"], end=w["end"],
                             speaker=speaker)
                    )
                return RecognizeResult(transcript=transcript, words=words)

    def stream(self):
        pass

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        pass

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        return AssemblyAiClient(token=key)



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import AudioException
from usttc.stream.stream import Stream
from usttc.asr_client.asr_provider import AsrProvider
from voicegain_speech import ApiClient
from voicegain_speech import Configuration
from voicegain_speech import TranscribeApi, DataApi
import time
import asyncio
import websockets
import threading


OFFLINE_FILE_SIZE_LIMIT = 150 * 1024 * 1024


class VoicegainClient(AsrClient):
    provider = AsrProvider.VOICEGAIN

    def __init__(self, api_client):
        self.api_client = api_client
        self.transcribe_api = TranscribeApi(api_client)
        self.data_api = DataApi(api_client)

    def recognize(self, audio: AudioFile, config: Config = Config()):
        if audio.file_size > OFFLINE_FILE_SIZE_LIMIT:
            raise AudioException("Audio file size is larger than 150MB. Voicegain does not support that")
        else:
            data_object = self.data_api.data_file_post(file=audio.file)
            object_id = data_object.object_id
            audio_source = {
                "dataStore": {
                    "uuid": object_id
                }
            }
            result = self._async_transcribe(config, audio, audio_source)
            self.data_api.data_delete(uuid=object_id)
            return result

    def stream(self, stream: Stream, config: Config):
        audio_ws_url, result_ws_url = self._start_async_web_socket_stream()
        result_ws_thread = threading.Thread(target=self._receive_result_thread, args=(result_ws_url, ))
        result_ws_thread.start()
        asyncio.run(self._stream_audio(stream, audio_ws_url))
        result_ws_thread.join()

    def _async_transcribe(self, config: Config, audio: AudioFile, audio_source):
        async_transcription_request = {
            "sessions": [
                {
                    "asyncMode": "OFF-LINE",
                    "poll": {
                        "afterlife": 60000,
                        "persist": 0
                    },
                    "content": {
                        "full": ["progress", "words", "transcript"]
                    }
                }
            ],
            "audio": {
                "source": audio_source
            }
        }

        if config.diarization:
            min_spk_count = config.diarization[0]
            max_spk_count = config.diarization[1]
            async_transcription_request["settings"] = {
                "asr": {
                    "diarization": {
                        "minSpeakers": min(max(1, min_spk_count), 10),
                        "maxSpeakers": min(max(1, max_spk_count), 10)
                    }
                }
            }

        def _get_result_resp(request):
            async_transcribe_init_response = self.transcribe_api.asr_transcribe_async_post(
                async_transcription_request=request
            )

            async_response_session = async_transcribe_init_response.sessions[0]
            session_id = async_response_session.session_id

            while True:
                time.sleep(5)
                poll_response = self.transcribe_api.asr_transcribe_async_get(
                    session_id=session_id,
                    full=True
                )
                poll_response_result = poll_response.result

                if poll_response_result.final:
                    return poll_response_result
                else:
                    pass

        if config.separate_speaker_per_channel and audio.channels == 2:
            async_transcription_request["sessions"][0]["audioChannelSelector"] = "left"
            result_resp_left = _get_result_resp(async_transcription_request)
            async_transcription_request["sessions"][0]["audioChannelSelector"] = "right"
            result_resp_right = _get_result_resp(async_transcription_request)
            result_responds = [result_resp_left, result_resp_right]
        else:
            result_responds = [_get_result_resp(async_transcription_request)]

        words = []
        for (M, result_resp) in enumerate(result_responds):
            if result_resp.words:
                for word in result_resp.words:
                    if len(result_responds) == 1:
                        spk = word.spk
                    else:
                        spk = M
                    words.append(Word(
                        text=word.utterance,
                        confidence=word.confidence,
                        start=word.start,
                        duration=word.duration,
                        speaker=spk
                    ))
        result = RecognizeResult(words=words)
        return result

    def _start_async_web_socket_stream(self):
        async_transcription_request = {
            "sessions": [
                {
                    "asyncMode": "REAL-TIME",
                    "websocket": {
                        "adHoc": True,
                        "useSTOMP": False,
                        "minimumDelay": 0
                    },
                    "content": {
                        "incremental": ['words'],
                        "full": []
                    }
                }
            ],
            "audio": {
                "source": {
                    "stream": {
                        "protocol": "WEBSOCKET"
                    }
                },
                "format": "L16",
                "channel": "mono",
                "rate": 16000
            }
        }
        async_transcribe_init_response = self.transcribe_api.asr_transcribe_async_post(
            async_transcription_request=async_transcription_request
        )
        audio_ws_url = async_transcribe_init_response.audio.stream.websocket_url
        result_ws_url = async_transcribe_init_response.sessions[0].websocket.url
        return audio_ws_url, result_ws_url

    @staticmethod
    async def _stream_audio(stream: Stream, audio_ws_url: str):
        async with websockets.connect(audio_ws_url,
                                      # we need to lower the buffer size - otherwise the sender will buffer for too long
                                      write_limit=480, compression=None) as websocket:
            try:
                for byte_buf in stream.generator():
                    await websocket.send(byte_buf)
                print("Waiting 5 seconds for processing to finish...", flush=True)
                time.sleep(5.0)
                print("done waiting", flush=True)
                await websocket.close()
            except Exception as e:
                print("Exception when sending audio via websocket: " + str(e))

    def _receive_result_thread(self, result_ws_url: str):
        try:
            asyncio.run(self._websocket_receive(result_ws_url))
        except Exception as e:
            print(e)

    @staticmethod
    async def _websocket_receive(result_ws_url: str):
        async with websockets.connect(result_ws_url,
                                      write_limit=480,
                                      # compression needs to be disabled otherwise will buffer for too long
                                      compression=None) as websocket:
            try:
                while True:
                    msg = await websocket.recv()
                    print(msg)
            except Exception as e:
                print(e)

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        with open(filename) as f:
            key = f.read().strip()
            return VoicegainClient.from_key(key, *args, **kwargs)

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        configuration = Configuration()
        configuration.access_token = key
        api_client = ApiClient(configuration=configuration)
        return VoicegainClient(api_client)



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile, AudioFormat
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import ConfigurationException, AudioException
from usttc.asr_client.asr_provider import AsrProvider
import time
import boto3
import requests
from usttc.utils import generate_random_str


AUDIO_DURATION_LIMIT = 4 * 60 * 60
FILE_SIZE_LIMIT = 2 * 1024 * 1024 * 1024


class AwsClient(AsrClient):
    provider = AsrProvider.AMAZON_AWS

    def __init__(self, s3_client, transcribe_client, s3_bucket):
        self.s3_client = s3_client
        self.transcribe_client = transcribe_client
        self.s3_bucket = s3_bucket

    def recognize(self, audio: AudioFile, config: Config = Config()):

        if audio.duration > AUDIO_DURATION_LIMIT:
            raise AudioException("AWS does not support audio longer than 4 hours")

        convert_audio = False
        if audio.codec not in {
            AudioFormat.LINEAR16, AudioFormat.FLAC, AudioFormat.MULAW, AudioFormat.AMR, AudioFormat.MP3, AudioFormat.MP4
        }:
            audio = audio.convert()
            convert_audio = True

        if audio.file_size > FILE_SIZE_LIMIT:
            if convert_audio:
                audio.delete()
            raise AudioException("AWS does not support audio larger than 2GB")

        s3_object_name = "{}{}".format(generate_random_str(20), audio.file_extension)
        self.s3_client.upload_file(
            audio.file, self.s3_bucket, s3_object_name
        )

        job_name = "job-{}".format(generate_random_str(20))
        job_uri = "s3://{}/{}".format(self.s3_bucket, s3_object_name)

        settings = {}
        if config.diarization:
            max_spk_count = config.diarization[1]
            settings["MaxSpeakerLabels"] = min(max(max_spk_count, 2), 10)
            settings["ShowSpeakerLabels"] = True
        if config.separate_speaker_per_channel and audio.channels > 1:
            settings["ChannelIdentification"] = True

        self.transcribe_client.start_transcription_job(
            TranscriptionJobName=job_name,
            Media={'MediaFileUri': job_uri},
            MediaFormat=audio.file_extension_no_dot,
            LanguageCode=config.language,
            Settings=settings
        )

        def _process_items(items, spk_id=None, _speaker_map=None):
            words = []
            for i in items:
                if i["type"] == "pronunciation":
                    alternatives = i["alternatives"]
                    if alternatives:
                        speaker_id = spk_id
                        if (speaker_id is None) and (_speaker_map is not None):
                            speaker_id = _speaker_map.get((i["start_time"], i["end_time"]))
                        word = Word(
                            text=alternatives[0]["content"],
                            confidence=alternatives[0]["confidence"],
                            start=float(i["start_time"]) * 1000,
                            end=float(i["end_time"]) * 1000,
                            speaker=speaker_id
                        )
                        words.append(word)
                elif (i["type"] == "punctuation") and words:
                    alternatives = i["alternatives"]
                    if alternatives:
                        content = alternatives[0]["content"]
                        words[-1].append_text(content)
            return words

        while True:
            status = self.transcribe_client.get_transcription_job(TranscriptionJobName=job_name)
            if status['TranscriptionJob']['TranscriptionJobStatus'] == 'COMPLETED':
                r = requests.get(status['TranscriptionJob']['Transcript']['TranscriptFileUri'])
                results = r.json()['results']
                word_list = []
                # multi-channel
                if ("channel_labels" in results) and ("channels" in results["channel_labels"]):
                    for channel in results["channel_labels"]["channels"]:
                        speaker = channel["channel_label"]
                        word_list += _process_items(channel["items"], spk_id=speaker)
                else:
                    # generate speaker map
                    speaker_map = dict()
                    if ("speaker_labels" in results) and ("segments" in results["speaker_labels"]):
                        for segment in results["speaker_labels"]["segments"]:
                            for item in segment["items"]:
                                speaker_map[(item["start_time"], item["end_time"])] = item["speaker_label"]
                    word_list += _process_items(results["items"], _speaker_map=speaker_map)
                break
            elif status['TranscriptionJob']['TranscriptionJobStatus'] == 'FAILED':
                word_list = None
                break
            time.sleep(5)
        self.transcribe_client.delete_transcription_job(TranscriptionJobName=job_name)
        self.s3_client.delete_object(Bucket=self.s3_bucket, Key=s3_object_name)
        if convert_audio:
            audio.delete()
        return RecognizeResult(words=word_list)

    def stream(self):
        pass

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        raise ConfigurationException("AWS ASR: Use key authentication")

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        aws_access_key_id = key
        aws_secret_access_key = kwargs.get("aws_secret_access_key")
        if not aws_secret_access_key:
            raise ConfigurationException("AWS ASR: Specify aws_secret_access_key arg")
        region_name = kwargs.get("region_name")
        if not region_name:
            raise ConfigurationException("AWS ASR: Specify region_name arg")
        s3_bucket = kwargs.get("s3_bucket")
        if not s3_bucket:
            raise ConfigurationException("AWS ASR: Specify s3_bucket arg")
        s3_client = boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name)
        transcribe_client = boto3.client(
            'transcribe',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name)
        return AwsClient(s3_client, transcribe_client, s3_bucket)



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import ConfigurationException
from usttc.asr_client.asr_provider import AsrProvider
from rev_ai import apiclient, JobStatus
import time


class RevClient(AsrClient):
    provider = AsrProvider.REV

    def __init__(self, client):
        self.client = client

    def recognize(self, audio: AudioFile, config: Config = Config()):
        speaker_channels_count = None
        if config.separate_speaker_per_channel and audio.channels > 1:
            speaker_channels_count = audio.channels
        job = self.client.submit_job_local_file(
            audio.file,
            skip_diarization=(config.diarization is None),
            speaker_channels_count=speaker_channels_count
        )

        while True:
            time.sleep(5)
            job_details = self.client.get_job_details(job.id)
            if job_details.status == JobStatus.TRANSCRIBED:
                json_result = self.client.get_transcript_json(job.id)
                # transcript = ""
                words = []
                for monologue in json_result["monologues"]:
                    speaker = monologue["speaker"]
                    for element in monologue["elements"]:
                        # transcript += element["value"]
                        if element["type"] == "text":
                            words.append(
                                Word(
                                    text=element["value"], confidence=element["confidence"],
                                    start=element["ts"] * 1000,
                                    end=element["end_ts"] * 1000,
                                    speaker=speaker
                                )
                            )
                        elif (element["type"] == "punct") and words:
                            value_strip = element["value"].strip()
                            if value_strip:
                                words[-1].append_text(value_strip)
                break
            elif job_details.status == JobStatus.FAILED:
                # transcript = ""
                words = None
                break

        self.client.delete_job(job.id)
        return RecognizeResult(words=words)

    def stream(self):
        pass

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        raise ConfigurationException("Rev.ai ASR: Use key authentication")

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        client = apiclient.RevAiAPIClient(key)
        return RevClient(client)



from usttc.asr_client.asr_client_interface import AsrClientInterface
from abc import abstractmethod


class AsrClient(AsrClientInterface):

    @staticmethod
    @abstractmethod
    def from_key_file(filename: str, *args, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def from_key(key: str, *args, **kwargs):
        pass



import re


class StreamResult:
    def __init__(self, transcript, is_final):
        self.transcript = transcript.strip()
        self.is_final = is_final


class StreamResults:
    def __init__(self, responses):
        self.stream_result_list = []
        self.responses = responses

    def get_transcript(self):
        return " ".join(i.transcript for i in self.stream_result_list if i.is_final)

    def till_end(self):
        for response in self.responses:
            if not response.is_final:
                if self.stream_result_list and (not self.stream_result_list[-1].is_final) and \
                        (self.stream_result_list[-1].transcript == response.transcript):
                    continue
                print("MID: " + response.transcript)
                self.stream_result_list.append(response)
            else:
                self.stream_result_list.append(response)
                print("FIN: " + response.transcript)

                # Exit recognition if any of the transcribed phrases could be
                # one of our keywords.
                if re.search(r"\b(exit|quit)\b", response.transcript, re.I):
                    print("Exiting..")
                    break



from usttc.asr_client.asr_client_interface import AsrClientInterface
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.stream.stream import Stream


class MultiAsrClient(AsrClientInterface):

    def __init__(self, asr_clients):
        """
        :param asr_clients: List of asr_client
        """
        self.asr_clients = asr_clients

    def recognize(self, audio: AudioFile, config: Config = Config()):
        results_map = dict()
        for asr_client in self.asr_clients:
            result = asr_client.recognize(audio, config)
            results_map[asr_client.provider] = result
        return results_map

    def stream(self, stream: Stream, config: Config):
        pass



from usttc.result.text_unit import TextUnit


class Word(TextUnit):
    def __init__(self, text, confidence=1, start=None, end=None, duration=None, speaker=None):
        super().__init__(text, confidence, start, end, duration, speaker)

    def __repr__(self):
        return "[Word] Speaker-{}: {} [{} : {}] (conf: {})".format(
            self.speaker, self.text, self.start, self.end, self.confidence
        )

    @property
    def speaker(self):
        if self._speaker is not None:
            return self._speaker
        return 0

    @speaker.setter
    def speaker(self, s):
        self._speaker = s



from usttc.result.paragraph import Paragraph


class RecognizeResult:
    def __init__(self, transcript=None, words=None):
        self._transcript = transcript
        if self._transcript is not None:
            self._transcript = self._transcript.strip()
        self._words = words
        self._dialogue = None
        self.post_process_done = False

    @property
    def transcript(self):
        if self._transcript:
            return self._transcript
        if self.words:
            self._transcript = " ".join([i.text for i in self.words])
            return self._transcript
        return ""

    @property
    def words(self):
        if self._words:
            if not self.post_process_done:
                self._post_process_word()
            return self._words
        return None

    @property
    def dialogue(self):
        def create_paragraph(word_list):
            text = " ".join([w.text for w in word_list])
            conf = 0
            for w in word_list:
                conf += w.confidence
            conf /= len(word_list)
            return Paragraph(
                text=text, confidence=conf,
                start=word_list[0].start, end=word_list[-1].end, speaker=word_list[0].speaker
            )

        if self._dialogue is None:
            self._dialogue = []
            current_para = []
            for word in self.words:
                if (not current_para) or (current_para[-1].speaker == word.speaker):
                    current_para.append(word)
                else:
                    self._dialogue.append(create_paragraph(current_para))
                    current_para = [word]
            if current_para:
                self._dialogue.append(create_paragraph(current_para))
        return self._dialogue

    @property
    def pretty_dialogue(self):
        dialogue = self.dialogue
        pretty_text_list = []
        for paragraph in dialogue:
            pretty_text_list.append("Speaker-{} : {}".format(paragraph.speaker, paragraph.text))
        return "\n".join(pretty_text_list)

    def _post_process_word(self):
        self._words.sort(key=lambda x: x.start)
        current_spk_id_map = dict()
        for word in self._words:
            speaker = word.speaker
            if speaker in current_spk_id_map:
                word.speaker = current_spk_id_map.get(speaker)
            else:
                new_id = len(current_spk_id_map)
                current_spk_id_map[speaker] = new_id
                word.speaker = new_id

        self.post_process_done = True



from abc import ABC, abstractmethod
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.stream.stream import Stream


class AsrClientInterface(ABC):

    @abstractmethod
    def recognize(self, audio: AudioFile, config: Config = Config()):
        pass

    @abstractmethod
    def stream(self, stream: Stream, config: Config):
        pass



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile, AudioFormat
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import ConfigurationException, AudioException
from usttc.asr_client.asr_provider import AsrProvider
import azure.cognitiveservices.speech as speechsdk
import json
import logging


class AzureClient(AsrClient):
    provider = AsrProvider.MICROSOFT_AZURE

    def __init__(self, key, region):
        self.key = key
        self.region = region

    def recognize(self, audio: AudioFile, config: Config = Config()):

        if config.diarization:
            raise ConfigurationException("Azure python SDK does not support diarization. "
                                         "Will switch to batch transcription API later on")
        if config.separate_speaker_per_channel and audio.channels > 1:
            raise ConfigurationException("Azure python SDK does not support multi-channel audio. "
                                         "Will switch to batch transcription API later on")

        speech_config = speechsdk.SpeechConfig(subscription=self.key,
                                               region=self.region)
        speech_config.request_word_level_timestamps()

        convert_audio = False
        if audio.codec != AudioFormat.LINEAR16:
            audio = audio.convert()
            convert_audio = True

        audio_input = speechsdk.AudioConfig(filename=audio.file)
        speech_recognizer = speechsdk.SpeechRecognizer(speech_config=speech_config, audio_config=audio_input)

        result = speech_recognizer.recognize_once_async().get()
        transcript = result.text
        n_best = json.loads(result.json).get("NBest")
        words = []
        if n_best:
            top_1 = n_best[0]
            confidence = top_1["Confidence"]
            for w in top_1["Words"]:
                words.append(
                    Word(
                        text=w["Word"], confidence=confidence,
                        start=w["Offset"] / 10000,
                        duration=w["Duration"] / 10000
                    )
                )

        if convert_audio:
            audio.delete()
        return RecognizeResult(transcript=transcript, words=words)

    def stream(self):
        pass

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        raise ConfigurationException("Azure ASR: Use key authentication")

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        logging.info("Azure integration is Alpha version")
        region = kwargs.get("region")
        if not region:
            raise ConfigurationException("Azure ASR: Specify region arg")
        return AzureClient(key, region)



from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import ConfigurationException, AudioException
from usttc.asr_client.asr_provider import AsrProvider
from deepgram import Deepgram
import asyncio


class DeepgramClient(AsrClient):
    provider = AsrProvider.DEEPGRAM

    def __init__(self, client):
        self.client = client

    def recognize(self, audio: AudioFile, config: Config = Config()):
        response = asyncio.run(self.async_recognize(config, audio))
        word_list = []
        results = response["results"]
        channels = results.get("channels")
        if channels:
            for (M, channel) in enumerate(channels):
                alternatives = channel["alternatives"]
                words = alternatives[0]["words"]
                for word in words:
                    if len(channels) > 1:
                        speaker = M
                    else:
                        speaker = word.get("speaker")
                    word_list.append(
                        Word(
                            text=word["punctuated_word"], confidence=word["confidence"],
                            start=word["start"] * 1000,
                            end=word["end"] * 1000,
                            speaker=speaker
                        )
                    )

        return RecognizeResult(words=word_list)

    async def async_recognize(self, config: Config, audio: AudioFile):
        with open(audio.file, 'rb') as audio_file:
            # Replace mimetype as appropriate
            source = {'buffer': audio_file, 'mimetype': 'audio/*'}
            response = await self.client.transcription.prerecorded(
                source,
                {
                    'punctuate': True,
                    "diarize": (config.diarization is not None),
                    "multichannel": config.separate_speaker_per_channel and (audio.channels > 1)
                }
            )
            return response

    def stream(self):
        pass

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        raise ConfigurationException("Deepgram ASR: Use key authentication")

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        client = Deepgram(key)
        return DeepgramClient(client)





class TextUnit:
    def __init__(self, text, confidence=1, start=None, end=None, duration=None, speaker=None):
        self._text = text
        self._confidence = float(confidence)
        self._speaker = speaker
        self._start = int(start)
        if (end is None) and (duration is None):
            raise ValueError("Both end and duration are None")
        if duration is None:
            self._end = int(end)
            self._duration = max(0, self._end - self._start)
        elif end is None:
            self._duration = int(duration)
            self._end = self._start + self._duration
        else:
            self._end = int(end)
            self._duration = int(duration)
            if self._start + self._duration != self._end:
                raise ValueError("start + duration != end")

    def append_text(self, new_text):
        self._text += new_text

    @property
    def text(self):
        return self._text

    @property
    def confidence(self):
        return self._confidence

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def duration(self):
        return self._duration

    @property
    def speaker(self):
        return self._speaker



import ffmpeg
from enum import Enum
import base64
import os
from usttc.utils import generate_random_str
from usttc.exceptions.exceptions import AudioException


class AudioFile:
    def __init__(self, file_path):
        self._file = file_path
        self._valid = False
        try:
            probe_dict = ffmpeg.probe(file_path)
            # print(probe_dict)
            self._duration = float(probe_dict["format"]["duration"])
            audio_stream = None
            for stream in probe_dict["streams"]:
                if stream.get("codec_type") == "audio":
                    audio_stream = stream
                    break
            if audio_stream:
                self._channels = int(audio_stream["channels"])
                if self._channels > 2:
                    AudioException("Only support mono or stereo audio")
                self._sample_rate = int(audio_stream["sample_rate"])
                self._codec = self._get_audio_format(audio_stream["codec_name"], self.file_extension_no_dot)
                if self._codec:
                    self._valid = True
        except AudioException as e:
            raise e
        except:
            pass
        if not self._valid:
            raise AudioException("Audio file is not valid")

    def convert(self, to="wav", sample_rate=16000):
        file_dir, _ = os.path.split(self._file)
        new_file_path = os.path.join(file_dir, generate_random_str(10) + ".{}".format(to))
        stream = ffmpeg.input(self._file)
        stream = ffmpeg.output(stream, new_file_path, **{'ar': str(sample_rate)})
        ffmpeg.run(stream, capture_stdout=True, capture_stderr=True)
        return AudioFile(file_path=new_file_path)

    def delete(self):
        os.remove(self._file)

    @property
    def file_size(self):
        return os.path.getsize(self._file)

    @property
    def file_extension(self):
        return os.path.splitext(self._file)[-1]

    @property
    def file_extension_no_dot(self):
        return self.file_extension.strip(".")

    @property
    def duration(self):
        return self._duration

    @property
    def channels(self):
        return self._channels

    @property
    def codec(self):
        return self._codec

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def byte_array_content(self):
        if self._file:
            with open(self._file, "rb") as f:
                return f.read()
        return None

    @property
    def base64_content(self):
        byte_array_content = self.byte_array_content
        if byte_array_content is not None:
            return base64.b64encode(byte_array_content).decode()
        return None

    @property
    def file(self):
        return self._file

    @staticmethod
    def _get_audio_format(codec_name, ext):
        format_map = {
            "pcm_s16le": AudioFormat.LINEAR16,
            "flac": AudioFormat.FLAC,
            "pcm_mulaw": AudioFormat.MULAW,
            "amr_nb": AudioFormat.AMR,
            "mp3": AudioFormat.MP3
        }
        if codec_name in format_map:
            return format_map[codec_name]
        if codec_name == "aac" and ext == "mp4":
            return AudioFormat.MP4
        return AudioFormat.OTHERS


class AudioFormat(Enum):
    OTHERS = 1
    LINEAR16 = 2
    FLAC = 3
    MULAW = 4
    AMR = 5
    MP3 = 6
    MP4 = 7



import setuptools

with open('requirements.txt', "r") as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="usttc",
    version="0.0.4",
    author="Kuo Zhang",
    author_email="kuo.zh92@gmail.com",
    description="Unified Speech-to-text Client",
    download_url='https://github.com/kakakuoka/usttc/archive/refs/tags/v0.0.4.tar.gz',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    packages=setuptools.find_packages(),
    include_package_data=True
)




class AsrBridgeException(Exception):
    pass


class AudioException(AsrBridgeException):
    pass


class ASRException(AsrBridgeException):
    pass


class ConfigurationException(AsrBridgeException):
    pass



from usttc.asr_client.asr_provider import AsrProvider
from usttc.asr_client.google_client import GoogleClient
from usttc.asr_client.voicegain_client import VoicegainClient
from usttc.asr_client.aws_client import AwsClient
from usttc.asr_client.assembly_ai_client import AssemblyAiClient
# from usttc.asr_client.azure_client import AzureClient
from usttc.asr_client.rev_client import RevClient
from usttc.asr_client.deepgram_client import DeepgramClient


class AsrClientFactory:

    @staticmethod
    def get_client_from_key_file(asr_provider: AsrProvider, filename: str, *args, **kwargs):
        return AsrClientFactory.get_provider_client(asr_provider).from_key_file(filename, *args, **kwargs)

    @staticmethod
    def get_client_from_key(asr_provider: AsrProvider, key: str, *args, **kwargs):
        return AsrClientFactory.get_provider_client(asr_provider).from_key(key, *args, **kwargs)

    @staticmethod
    def get_provider_client(asr_provider: AsrProvider):
        provider_map = {
            AsrProvider.GOOGLE: GoogleClient,
            AsrProvider.VOICEGAIN: VoicegainClient,
            AsrProvider.AMAZON_AWS: AwsClient,
            AsrProvider.REV: RevClient,
            AsrProvider.DEEPGRAM: DeepgramClient,
            AsrProvider.ASSEMBLY_AI: AssemblyAiClient,
            # AsrProvider.MICROSOFT_AZURE: AzureClient,
        }
        return provider_map.get(asr_provider)




from usttc.asr_client.asr_client import AsrClient
from usttc.config import Config
from usttc.audio.audio_file import AudioFile, AudioFormat
from usttc.result.recognize_result import RecognizeResult
from usttc.result.word import Word
from usttc.exceptions.exceptions import ConfigurationException, AudioException
from usttc.utils import generate_random_str
from usttc.stream.stream_results import StreamResult, StreamResults
from usttc.stream.stream import Stream
from google.cloud import speech
from google.cloud import storage
from usttc.asr_client.asr_provider import AsrProvider

AUDIO_DURATION_LIMIT = 480 * 60


class GoogleClient(AsrClient):
    provider = AsrProvider.GOOGLE

    def __init__(self, client, storage_client, google_storage_bucket):
        self.client = client
        self.storage_client = storage_client
        self.google_storage_bucket = google_storage_bucket

    def recognize(self, audio: AudioFile, config: Config = Config()):
        if audio.duration >= AUDIO_DURATION_LIMIT:
            raise AudioException("Google does not support audio longer than 480 minutes")

        convert_audio = False
        if (audio.codec not in {AudioFormat.LINEAR16, AudioFormat.FLAC, AudioFormat.MULAW, AudioFormat.AMR}) or \
            ((audio.codec not in {AudioFormat.LINEAR16, AudioFormat.FLAC})
                and (audio.channels > 1)
                and config.separate_speaker_per_channel):
            audio = audio.convert()
            convert_audio = True

        if audio.duration < 60:
            recog_audio = speech.RecognitionAudio(
                content=audio.byte_array_content
            )
            result = self._async_recognize(config, audio, recog_audio)
        else:
            blob = self._upload_to_google_storage(self.google_storage_bucket, audio)
            gs_url = "gs://{}/{}".format(blob.bucket.name, blob.name)
            recog_audio = speech.RecognitionAudio()
            recog_audio.uri = gs_url
            result = self._async_recognize(config, audio, recog_audio)
            blob.delete()

        if convert_audio:
            audio.delete()
        return result

    def stream(self, stream: Stream, config: Config):

        recog_config = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.LINEAR16,
            sample_rate_hertz=16000,
            language_code=config.language,
        )
        streaming_config = speech.StreamingRecognitionConfig(
            config=recog_config,
            interim_results=True
        )

        audio_generator = stream.generator()
        requests = (
            speech.StreamingRecognizeRequest(audio_content=content)
            for content in audio_generator
        )
        responses = self.client.streaming_recognize(streaming_config, requests)

        def parse_response():
            for response in responses:
                if not response.results:
                    continue
                result = response.results[0]
                if not result.alternatives:
                    continue
                transcript = result.alternatives[0].transcript
                if not result.is_final:
                    yield StreamResult(transcript, False)
                else:
                    yield StreamResult(transcript, True)

        return StreamResults(parse_response())

    def _async_recognize(self, config: Config, audio: AudioFile, recog_audio):
        recognition_config = speech.RecognitionConfig(
            encoding=audio.codec.name,
            sample_rate_hertz=audio.sample_rate,
            language_code=config.language,
            model="video",
            use_enhanced=True,
            enable_word_time_offsets=True,
            enable_automatic_punctuation=True
        )

        if config.diarization:
            min_spk_count = config.diarization[0]
            max_spk_count = config.diarization[1]
            diarization_config = speech.SpeakerDiarizationConfig(
                enable_speaker_diarization=True, min_speaker_count=min_spk_count, max_speaker_count=max_spk_count
            )
            recognition_config.diarization_config = diarization_config

        if config.separate_speaker_per_channel and audio.channels > 1:
            recognition_config.enable_separate_recognition_per_channel = True
            recognition_config.audio_channel_count = audio.channels

        operation = self.client.long_running_recognize(config=recognition_config, audio=recog_audio)
        response = operation.result()
        words = []
        for current_result in response.results:
            current_channel_tag = current_result.channel_tag
            for w in current_result.alternatives[0].words:
                start = w.start_time.total_seconds() * 1000
                end = w.end_time.total_seconds() * 1000
                if current_channel_tag:
                    speaker = current_channel_tag
                else:
                    speaker = w.speaker_tag
                words.append(Word(text=w.word, start=start, end=end, speaker=speaker))

        return RecognizeResult(transcript=None, words=words)

    def _upload_to_google_storage(self, bucket_name, audio: AudioFile):
        bucket = self.storage_client.bucket(bucket_name)
        blob = bucket.blob("{}{}".format(generate_random_str(20), audio.file_extension))
        blob.upload_from_filename(audio.file)
        return blob

    @staticmethod
    def from_key_file(filename: str, *args, **kwargs):
        google_storage_bucket = kwargs.get("google_storage_bucket")
        if not google_storage_bucket:
            raise ConfigurationException("Google ASR: Specify google_storage_bucket arg")
        client = speech.SpeechClient.from_service_account_file(filename)
        storage_client = storage.Client.from_service_account_json(filename)
        return GoogleClient(client, storage_client, google_storage_bucket)

    @staticmethod
    def from_key(key: str, *args, **kwargs):
        raise ConfigurationException("Google ASR: Use key file authentication")



import pyaudio
from six.moves import queue
from usttc.stream.stream import Stream
import time


class MicrophoneStream(Stream):
    """Opens a recording stream as a generator yielding the audio chunks."""

    def __init__(self, rate=16000):
        self._rate = rate
        self._chunk = int(rate/100)  # 10ms

        # Create a thread-safe buffer of audio data
        self._buff = queue.Queue()
        self.closed = True

    def __enter__(self):
        self._audio_interface = pyaudio.PyAudio()
        self._audio_stream = self._audio_interface.open(
            format=pyaudio.paInt16,
            # The API currently only supports 1-channel (mono) audio
            # https://goo.gl/z757pE
            channels=1,
            rate=self._rate,
            input=True,
            frames_per_buffer=self._chunk,
            # Run the audio stream asynchronously to fill the buffer object.
            # This is necessary so that the input device's buffer doesn't
            # overflow while the calling thread makes network requests, etc.
            stream_callback=self._fill_buffer,
        )
        self.closed = False
        return self

    def __exit__(self, type, value, traceback):
        self._audio_stream.stop_stream()
        self._audio_stream.close()
        self.closed = True
        # Signal the generator to terminate so that the client's
        # streaming_recognize method will not block the process termination.
        self._buff.put(None)
        self._audio_interface.terminate()

    def _fill_buffer(self, in_data, frame_count, time_info, status_flags):
        """Continuously collect data from the audio stream, into the buffer."""
        self._buff.put(in_data)
        return None, pyaudio.paContinue

    def generator(self):
        while not self.closed:
            # Use a blocking get() to ensure there's at least one chunk of
            # data, and stop iteration if the chunk is None, indicating the
            # end of the audio stream.
            chunk = self._buff.get()
            if chunk is None:
                return
            data = [chunk]

            # Now consume whatever other data's still buffered.
            while True:
                try:
                    chunk = self._buff.get(block=False)
                    if chunk is None:
                        return
                    data.append(chunk)
                except queue.Empty:
                    break

            yield b"".join(data)
